{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\n# Distribution\n\nThis section introduces the usage of the distributed mode of AgentScope. AgentScope natively provides a distributed mode based on gRPC.\nIn this mode, multiple agents in one application can be deployed to different processes or even different machines, thereby fully utilizing computational resources and improving efficiency.\n\n## Basic Usage\n\nThe distributed mode of AgentScope requires no modification to the main\nprocess code compared to the traditional mode. You only need to call the\n``to_dist`` function when initializing your agent.\n\nThis section will showcase how to use the distributed mode of AgentScope with\na web search case.\nTo highlight the acceleration effect brought by the distributed mode of\nAgentScope, a simple custom ``WebAgent`` is used here.\n``WebAgent`` will sleep 5 seconds to simulate the process of crawling a\nwebpage and looking for answers from it, and there are 5 agents in the example.\n\nThe process of performing the search is the ``run`` function. The difference\nbetween the traditional and distributed mode is only in the initialization\nstage, namely ``init_without_dist`` and ``init_with_dist``.\nIn distributed mode, you have to call the ``to_dist`` function, which turns\nthe original Agent into the corresponding distributed version.\n\n```python\n# Please do not run this code in jupyter notebook.\n# Please copy the code to the ``dist_main.py`` file and use the ``python dist_main.py`` command to run this code.\n# Please install the distributed version of agentscope before running this code.\n\nimport time\nimport agentscope\nfrom agentscope.agents import AgentBase\nfrom agentscope.message import Msg\n\nclass WebAgent(AgentBase):\n\n    def __init__(self, name):\n        super().__init__(name)\n\n    def get_answer(self, url: str, query: str):\n        time.sleep(5)\n        return f\"Answer from {self.name}\"\n\n    def reply(self, x: dict = None) -> dict:\n        return Msg(\n            name=self.name,\n            content=self.get_answer(x.content[\"url\"], x.content[\"query\"])\n        )\n\n\nQUERY = \"example query\"\nURLS = [\"page_1\", \"page_2\", \"page_3\", \"page_4\", \"page_5\"]\n\ndef init_without_dist():\n    return [WebAgent(f\"W{i}\") for i in range(len(URLS))]\n\n\ndef init_with_dist():\n    return [WebAgent(f\"W{i}\").to_dist() for i in range(len(URLS))]\n\n\ndef run(agents):\n    start = time.time()\n    results = []\n    for i, url in enumerate(URLS):\n        results.append(agents[i].reply(\n            Msg(\n                name=\"system\",\n                role=\"system\",\n                content={\n                    \"url\": url,\n                    \"query\": QUERY\n                }\n            )\n        ))\n    for result in results:\n        print(result.content)\n    end = time.time()\n    return end - start\n\n\nif __name__ == \"__main__\":\n    agentscope.init()\n    start = time.time()\n    simple_agents = init_without_dist()\n    dist_agents = init_with_dist()\n    end = time.time()\n    print(f\"Initialization time: {end - start}\")\n    print(f\"Runtime without distributed mode: {run(simple_agents)}\")\n    print(f\"Runtime with distributed mode: {run(dist_agents)}\")\n```\nThe output of running this sample is as follows:\n\n```text\nInitialization time: 16.50428819656372\nAnswer from W0\nAnswer from W1\nAnswer from W2\nAnswer from W3\nAnswer from W4\nRuntime without distributed mode: 25.034368991851807\nAnswer from W0\nAnswer from W1\nAnswer from W3\nAnswer from W2\nAnswer from W4\nRuntime with distributed mode: 5.0517587661743164\n```\nFrom the sample output above, we can observe that the running speed has\nsignificantly improved after adopting the distributed mode (25 s -> 5 s).\n\nThe example above is the most common use case for the distributed mode of\nAgentScope. It is recommended to use this method directly when not aiming\nfor extreme performance. If you require further performance optimization,\na deeper understanding of the distributed mode of AgentScope is necessary.\nThe advanced usage method of AgentScope distributed mode will be introduced\nin the following sections.\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Advanced Usage\n\nThis section will introduce advanced usage methods for the AgentScope distributed model to further improve operational efficiency.\n\n### Basic Concepts\n\n\nBefore diving into the advanced usage, we must first cover some basic concepts of the AgentScope distributed model.\n\n- **Main Process**: The process where the AgentScope application resides is referred to as the main process. For example, the ``run`` function in the previous section runs in the main process. There will only be one main process in each AgentScope application.\n- **Agent Server Process**: The AgentScope agent server process is where the agent runs in distributed mode. For example, in the previous section, all agents in ``dist_agents`` run in the agent server process. There can be multiple AgentScope agent server processes. These processes can run on any network-reachable machine and numerous agents can run simultaneously within each agent server process.\n- **Child Mode**: In child mode, the agent server process is started by the main process as a child process. For example, in the previous section, each agent in ``dist_agents`` is actually a child process of the main process. This mode is the default mode, which means if you call the ``to_dist`` function without any parameters, it will use this mode.\n- **Independent Mode**: In independent mode, the agent servers are independent of the main process and need to be pre-started on the machine. Specific parameters must be passed to the ``to_dist`` function, and the usage method will be introduced in the following section.\n\n### Using Independent Mode\n\nCompared to child mode, independent mode can avoid the overhead of initializing child processes, thus reducing the delay at the beginning of execution. This can effectively improve efficiency for programs that make multiple calls to ``to_dist``.\n\nIn independent mode, you need to pre-start the agent server processes on the machine and pass specific parameters to the ``to_dist`` function. Here, we will continue using the example from the basic usage section, assuming the code file for the basic usage is ``dist_main.py``. Then, create and run the following script separately.\n\n```python\n# Please do not run this code in a jupyter notebook.\n# Copy this code into a file named ``dist_server.py`` and run with the command ``python dist_server.py``.\n# Please install the distributed version of agentscope before running this code.\n# pip install agentscope[distributed]\n\nfrom dist_main import WebAgent\nimport agentscope\n\nif __name__ == \"__main__\":\n    agentscope.init()\n    assistant_server_launcher = RpcAgentServerLauncher(\n        host=\"localhost\",\n        port=12345,\n        custom_agent_classes=[WebAgent],\n    )\n    assistant_server_launcher.launch(in_subprocess=False)\n    assistant_server_launcher.wait_until_terminate()\n```\nThis script starts the AgentScope agent server process in the ``dist_server.py`` file, which is located in the same directory as the ``dist_main.py`` file from the basic usage section. Also, we need to make some minor modifications to the ``dist_main.py`` file by adding a new ``init_with_dist_independent`` function and replacing the call to ``init_with_dist`` with this new function.\n\n```python\ndef init_with_dist_independent():\n    return [WebAgent(f\"W{i}\").to_dist(host=\"localhost\", port=12345) for i in range(len(URLS))]\n\nif __name__ == \"__main__\":\n    agentscope.init()\n    start = time.time()\n    simple_agents = init_without_dist()\n    dist_agents = init_with_dist_independent()\n    end = time.time()\n    print(f\"Time taken for initialization: {end - start}\")\n    print(f\"Time taken without distributed mode: {run(simple_agents)}\")\n    print(f\"Time taken with distributed mode: {run(dist_agents)}\")\n```\nAfter completing the modifications, open a command prompt and run the ``dist_server.py`` file. Once it is successfully started, open another command prompt and run the ``dist_main.py`` file.\n\nAt this point, the initialization time in the output of ``dist_main.py`` will be significantly reduced. For example, the time taken here is only 0.02 seconds.\n\n```text\nTime taken for initialization: 0.018129825592041016\n...\n```\nIt's important to note that the above example uses ``host=\"localhost\"`` and ``port=12345``, and both ``dist_main.py`` and ``dist_server.py`` are running on the same machine. In actual usage, ``dist_server.py`` can run on a different machine. In this case, ``host`` should be set to the IP address of the machine running ``dist_server.py``, and ``port`` should be set to any available port, ensuring that different machines can communicate over the network.\n\n### Avoid Duplicate Initialization\n\nIn the code above, the ``to_dist`` function is called on an already initialized agent. The essence of ``to_dist`` is to clone the original agent to the agent server process, while retaining an ``RpcAgent`` in the main process as a proxy of the original agent. Calls to this ``RpcAgent`` will be forwarded to the corresponding agent in the agent server process.\n\nThere is a potential issue with this approach: the original Agent is initialized twice\u2014once in the main process and once in the agent server process\u2014and these initializations are executed sequentially, which cannot be accelerated via parallelism. For Agents with low initialization costs, directly calling the ``to_dist`` function won't significantly affect performance. However, for agents with high initialization costs, it is important to avoid redundant initialization. Therefore, the AgentScope distributed mode offers an alternative method for distributed mode initialization, which allows directly passing the ``to_dist`` parameter within any Agent's initialization function, as shown in the modified example below:\n\n```python\ndef init_with_dist():\n    return [WebAgent(f\"W{i}\", to_dist=True) for i in range(len(URLS))]\n\n\ndef init_with_dist_independent():\n    return [WebAgent(f\"W{i}\", to_dist={\"host\": \"localhost\", \"port\": \"12345\"}) for i in range(len(URLS))]\n```\nFor the subprocess mode, you only need to pass ``to_dist=True`` in the initialization function. For the independent process mode, you need to pass the parameters that were originally passed to the ``to_dist`` function as a dictionary to the ``to_dist`` field.\n\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.9.21"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}